"""
Creates a basic wrapper around a Piccolo model, turning it into an ASGI app.
"""
from __future__ import annotations
from datetime import timedelta
from functools import partial
import os
import typing as t

from piccolo.table import Table
from piccolo.apps.user.tables import BaseUser

from piccolo_admin import __VERSION__ as piccolo_admin_version
from piccolo_api.csrf.middleware import CSRFMiddleware
from piccolo_api.crud.endpoints import PiccoloCRUD
from piccolo_api.rate_limiting.middleware import (
    RateLimitingMiddleware,
    RateLimitProvider,
    InMemoryLimitProvider,
)
from piccolo_api.session_auth.endpoints import session_login, session_logout
from piccolo_api.session_auth.tables import SessionsBase
from piccolo_api.session_auth.middleware import SessionsAuthBackend

from starlette.responses import HTMLResponse, JSONResponse
from starlette.requests import Request
from starlette.routing import Router, Route, BaseRoute, Mount
from starlette.staticfiles import StaticFiles
from starlette.exceptions import ExceptionMiddleware
from starlette.middleware.authentication import AuthenticationMiddleware


ASSET_PATH = os.path.join(os.path.dirname(__file__), "dist")


def handle_auth_exception(request: Request, exc: Exception):
    return JSONResponse({"error": "Auth failed"}, status_code=401)


class AdminRouter(Router):
    """
    The root returns a single page app. The other URLs are REST endpoints.
    """

    table: t.List[Table] = []
    auth_table: t.Type[BaseUser] = BaseUser
    template: str = ""

    def __init__(
        self,
        *tables: t.Type[Table],
        auth_table: t.Type[BaseUser] = BaseUser,
        session_table: t.Type[SessionsBase] = SessionsBase,
        session_expiry: timedelta = timedelta(hours=1),
        max_session_expiry: timedelta = timedelta(days=7),
        increase_expiry: t.Optional[timedelta] = timedelta(minutes=20),
        page_size: int = 15,
        read_only: bool = False,
        rate_limit_provider: t.Optional[RateLimitProvider] = None,
        production: bool = False,
    ) -> None:
        self.auth_table = auth_table

        with open(os.path.join(ASSET_PATH, "index.html")) as f:
            self.template = f.read()

        auth_middleware = partial(
            AuthenticationMiddleware,
            backend=SessionsAuthBackend(
                auth_table=auth_table,
                session_table=session_table,
                admin_only=True,
                increase_expiry=increase_expiry,
            ),
            on_error=handle_auth_exception,
        )

        if not rate_limit_provider:
            rate_limit_provider = InMemoryLimitProvider(
                limit=1000, timespan=300
            )

        table_routes: t.List[BaseRoute] = [
            Mount(
                path=f"/{table._meta.tablename}/",
                app=PiccoloCRUD(
                    table, read_only=read_only, page_size=page_size
                ),
            )
            for table in tables
        ]
        table_routes += [
            Route(path="/", endpoint=self.get_table_list, methods=["GET"],)
        ]

        routes: t.List[BaseRoute] = [
            Route(path="/", endpoint=self.get_root, methods=["GET"]),
            Mount(
                path="/css",
                app=StaticFiles(directory=os.path.join(ASSET_PATH, "css")),
            ),
            Mount(
                path="/js",
                app=StaticFiles(directory=os.path.join(ASSET_PATH, "js")),
            ),
            Mount(
                path="/api",
                app=Router(
                    [
                        Mount(
                            path="/tables/",
                            app=auth_middleware(Router(table_routes)),
                        ),
                        Route(
                            path="/login/",
                            endpoint=RateLimitingMiddleware(
                                app=session_login(
                                    auth_table=self.auth_table,
                                    session_table=session_table,
                                    session_expiry=session_expiry,
                                    max_session_expiry=max_session_expiry,
                                    redirect_to=None,
                                    production=production,
                                ),
                                provider=rate_limit_provider,
                            ),
                            methods=["POST"],
                        ),
                        Route(
                            path="/logout/",
                            endpoint=session_logout(
                                session_table=session_table
                            ),
                            methods=["POST"],
                        ),
                        Mount(
                            path="/user/",
                            app=auth_middleware(
                                Router(
                                    [Route(path="/", endpoint=self.get_user)]
                                )
                            ),
                        ),
                        Mount(
                            path="/meta/",
                            app=auth_middleware(
                                Router(
                                    [Route(path="/", endpoint=self.get_meta)]
                                )
                            ),
                        ),
                    ]
                ),
            ),
        ]

        self.tables = tables
        super().__init__(routes)

    async def get_root(self, request: Request) -> HTMLResponse:
        return HTMLResponse(self.template)

    ###########################################################################

    def get_user(self, request: Request) -> JSONResponse:
        return JSONResponse(
            {
                "username": request.user.display_name,
                "user_id": request.user.user_id,
            }
        )

    ###########################################################################

    def get_meta(self, request: Request) -> JSONResponse:
        return JSONResponse({"piccolo_admin_version": piccolo_admin_version})

    ###########################################################################

    def get_table_list(self, request: Request) -> JSONResponse:
        return JSONResponse([i._meta.tablename for i in self.tables])


def get_all_tables(
    tables: t.Sequence[t.Type[Table]],
) -> t.Sequence[t.Type[Table]]:
    """
    Fetch any related tables, and include them.
    """
    output: t.List[t.Type[Table]] = []

    def get_references(table: t.Type[Table]):
        references = [
            i._foreign_key_meta.references
            for i in table._meta.foreign_key_columns
        ]
        for reference in references:
            if reference not in output:
                output.append(reference)
                get_references(reference)

    for table in tables:
        if table not in output:
            output.append(table)
        get_references(table)

    return output


def create_admin(
    tables: t.Sequence[t.Type[Table]],
    auth_table: t.Type[BaseUser] = BaseUser,
    session_table: t.Type[SessionsBase] = SessionsBase,
    session_expiry: timedelta = timedelta(hours=1),
    max_session_expiry: timedelta = timedelta(days=7),
    increase_expiry: t.Optional[timedelta] = timedelta(minutes=20),
    page_size: int = 15,
    read_only: bool = False,
    rate_limit_provider: t.Optional[RateLimitProvider] = None,
    production: bool = False,
    auto_include_related: bool = True,
):
    """
    :param tables:
        Each of the tables will be added to the admin.
    :param auth_table:
        Either a BaseUser, or BaseUser subclass table, which is used for
        fetching users.
    :param session_table:
        Either a SessionBase, or SessionBase subclass table, which is used
        for storing and querying session tokens.
    :param session_expiry:
        How long a session is valid for.
    :param max_session_expiry:
        The maximum time a session is valid for, taking into account any
        refreshes using `increase_expiry`.
    :param increase_expiry:
        If set, the `session_expiry` will be increased by this amount if it's
        close to expiry.
    :param page_size:
        The admin API paginates contents - this sets the number of results
        in each page.
    :param read_only:
        If True, all non auth endpoints only respond to GET requests - the
        admin can still be viewed, and the data can be filtered. Useful for
        creating online demos.
    :param rate_limit_provider:
        Rate limiting middleware is used to protect the login endpoint
        against brute force attack. If not set, an InMemoryLimitProvider
        will be configured with reasonable defaults.
    :param production:
        If True, the admin will enforce stronger security - for example,
        the cookies used will be secure, meaning they are only sent over
        HTTPS.
    :param auto_include_related:
        If a table has foreign keys to other tables, those tables will also be
        included in the admin by default, if not already specified. Otherwise
        the admin won't work as expected.
    """

    if auto_include_related:
        tables = get_all_tables(tables)

    return ExceptionMiddleware(
        CSRFMiddleware(
            AdminRouter(
                *tables,
                auth_table=auth_table,
                session_table=session_table,
                session_expiry=session_expiry,
                max_session_expiry=max_session_expiry,
                increase_expiry=increase_expiry,
                page_size=page_size,
                read_only=read_only,
                rate_limit_provider=rate_limit_provider,
                production=production,
            ),
        )
    )



"""
Runs the admin in read only mode - useful for letting people evaluate the admin
online without risk of abuse.
"""
import sys

import uvicorn
from piccolo_admin.example import (
    Director,
    Movie,
    Sessions,
    User,
    create_admin,
    populate_data,
)


APP = create_admin(
    [Director, Movie], auth_table=User, session_table=Sessions, read_only=True
)


def main(host="localhost", port=8080):
    populate_data()
    uvicorn.run(APP, host=host, port=port)


if __name__ == "__main__":
    args = sys.argv
    kwargs = {}
    if "--production" in args:
        kwargs.update({"host": "0.0.0.0", "port": 80})

    main(**kwargs)



import datetime


DIRECTORS = [
    {"id": 1, "name": "Peter Jackson"},
    {"id": 2, "name": "George Lucas"},
    {"id": 3, "name": "J.J. Abrams"},
    {"id": 4, "name": "Gareth Edwards"},
    {"id": 5, "name": "Irvin Kershner"},
    {"id": 6, "name": "Richard Marquand"},
    {"id": 7, "name": "Rian Johnson"},
    {"id": 8, "name": "Ron Howard"},
]

MOVIES = [
    {
        "name": "Star Wars: Episode IV - A New Hope",
        "rating": 93,
        "director": 2,
        "description": (
            "After Princess Leia, the leader of the Rebel Alliance, is held "
            "hostage by Darth Vader, Luke and Han Solo must free her and "
            "destroy the powerful weapon created by the Galactic Empire."
        ),
        "release_date": datetime.datetime(year=1977, month=12, day=27),
        "won_oscar": True,
    },
    {
        "name": "The Empire Strikes Back",
        "rating": 94,
        "director": 5,
        "description": (
            "Darth Vader is adamant about turning Luke Skywalker to the dark "
            "side. Master Yoda trains Luke to become a Jedi Knight while his "
            "friends try to fend off the Imperial fleet."
        ),
        "release_date": datetime.datetime(year=1980, month=5, day=20),
        "won_oscar": True,
    },
    {
        "name": "Return of the Jedi",
        "rating": 82,
        "director": 6,
        "description": (
            "Luke Skywalker attempts to bring his father back to the light "
            "side of the 'Force'. At the same time, the rebels hatch a plan "
            "to destroy the second Death Star."
        ),
        "release_date": datetime.datetime(year=1983, month=6, day=2),
        "won_oscar": True,
    },
    {
        "name": "Star Wars: Episode I – The Phantom Menace",
        "rating": 53,
        "director": 2,
        "description": (
            "Two Jedi Knights set out to search for someone who can "
            "bring peace to the Force. Their search ends when they come "
            "across a young, gifted boy. But the Sith returns to stake claim "
            "to the Force."
        ),
        "release_date": datetime.datetime(year=1999, month=7, day=16),
        "won_oscar": False,
    },
    {
        "name": "Star Wars: Episode II – Attack of the Clones",
        "rating": 65,
        "director": 2,
        "description": (
            "While pursuing an assassin, Obi Wan uncovers a sinister plot to "
            "destroy the Republic. With the fate of the galaxy hanging in the "
            "balance, the Jedi must defend the galaxy against the evil Sith."
        ),
        "release_date": datetime.datetime(year=2002, month=5, day=16),
        "won_oscar": False,
    },
    {
        "name": "Star Wars: Episode III – Revenge of the Sith",
        "rating": 80,
        "director": 2,
        "description": (
            "Anakin joins forces with Obi-Wan and sets Palpatine free from "
            "the evil clutches of Count Doku. But he falls prey to mind games "
            "that are played by Palpatine and the Jedis and gives into "
            "temptation."
        ),
        "release_date": datetime.datetime(year=2005, month=5, day=19),
        "won_oscar": False,
    },
    {
        "name": "Star Wars: The Force Awakens",
        "rating": 93,
        "director": 3,
        "description": (
            "A new order threatens to destroy the New Republic. Finn, Rey and "
            "Poe, backed by the Resistance and the Republic, must find a way "
            "to stop them and find Luke, the last surviving Jedi."
        ),
        "release_date": datetime.datetime(year=2015, month=12, day=17),
        "won_oscar": False,
    },
    {
        "name": "Rogue One: A Star Wars Story",
        "rating": 83,
        "director": 4,
        "description": (
            "When she's a child, Jyn's father is forcibly taken by the Empire "
            "to help them complete the Death Star. As an adult, she joins a "
            "group of resistance fighters, who aim to steal its blueprints."
        ),
        "release_date": datetime.datetime(year=2016, month=12, day=13),
        "won_oscar": False,
    },
    {
        "name": "Star Wars: The Last Jedi",
        "rating": 91,
        "director": 7,
        "description": (
            "Rey seeks to learn the ways of the Jedi under Luke Skywalker, "
            "its remaining member, to reinvigorate the Resistance's war "
            "against the First Order."
        ),
        "release_date": datetime.datetime(year=2017, month=12, day=14),
        "won_oscar": False,
    },
    {
        "name": "Solo: A Star Wars Story",
        "rating": 70,
        "director": 8,
        "description": (
            "In a galaxy where hyperfuel is in high demand, Han Solo gets "
            "involved in a large-scale heist within the criminal underworld "
            "and meets individuals who change his life."
        ),
        "release_date": datetime.datetime(year=2018, month=5, day=25),
        "won_oscar": False,
    },
    {
        "name": "Star Wars: The Rise of Skywalker",
        "rating": 54,
        "director": 3,
        "description": (
            "The surviving Resistance faces the First Order once more as Rey, "
            "Finn and Poe Dameron's journey continues. With the power and "
            "knowledge of generations behind them, the final battle commences."
        ),
        "release_date": datetime.datetime(year=2019, month=12, day=19),
        "won_oscar": False,
    },
    {
        "name": "The Lord of the Rings: The Fellowship of the Ring",
        "rating": 91,
        "director": 1,
        "description": (
            "A young hobbit, Frodo, who has found the One Ring that belongs "
            "to the Dark Lord Sauron, begins his journey with eight "
            "companions to Mount Doom, the only place where it can be "
            "destroyed."
        ),
        "release_date": datetime.datetime(year=2001, month=12, day=10),
        "won_oscar": True,
    },
    {
        "name": "The Lord of the Rings: The Two Towers",
        "rating": 95,
        "director": 1,
        "description": (
            "Frodo and Sam arrive in Mordor with the help of Gollum. A number "
            "of new allies join their former companions to defend Isengard as "
            "Saruman launches an assault on it."
        ),
        "release_date": datetime.datetime(year=2002, month=12, day=11),
        "won_oscar": True,
    },
    {
        "name": "The Lord of the Rings - The Return of the King",
        "rating": 93,
        "director": 1,
        "description": (
            "The former Fellowship members prepare for the final battle. "
            "While Frodo and Sam approach Mount Doom to destroy the One Ring, "
            "they follow Gollum unaware of the path he is leading them to."
        ),
        "release_date": datetime.datetime(year=2003, month=12, day=17),
        "won_oscar": True,
    },
    {
        "name": "The Hobbit: An Unexpected Journey",
        "rating": 64,
        "director": 1,
        "description": (
            "Bilbo Baggins, a hobbit, is persuaded into accompanying a wizard "
            "and a group of dwarves on a journey to reclaim the city of "
            "Erebor and all its riches from the dragon Smaug."
        ),
        "release_date": datetime.datetime(year=2012, month=12, day=12),
        "won_oscar": False,
    },
    {
        "name": "The Hobbit: The Desolation of Smaug",
        "rating": 74,
        "director": 1,
        "description": (
            "Bilbo Baggins, a hobbit, and his companions face great dangers "
            "on their journey to Laketown. Soon, they reach the Lonely "
            "Mountain, where Bilbo comes face-to-face with the fearsome "
            "dragon Smaug."
        ),
        "release_date": datetime.datetime(year=2013, month=12, day=13),
        "won_oscar": False,
    },
    {
        "name": "The Hobbit: The Battle of the Five Armies",
        "rating": 59,
        "director": 1,
        "description": (
            "Bilbo fights against a number of enemies to save the life of his "
            "Dwarf friends and protects the Lonely Mountain after a conflict "
            "arises."
        ),
        "release_date": datetime.datetime(year=2014, month=12, day=1),
        "won_oscar": False,
    },
]



"""
Import all of the Tables subclasses in your app here, and register them with
the APP_CONFIG.
"""

import os

from piccolo.conf.apps import AppConfig


CURRENT_DIRECTORY = os.path.dirname(os.path.abspath(__file__))


APP_CONFIG = AppConfig(
    app_name="piccolo_admin",
    migrations_folder_path=os.path.join(
        CURRENT_DIRECTORY, "piccolo_migrations"
    ),
    table_classes=[],
    migration_dependencies=[
        "piccolo_api.session_auth.piccolo_app",
        "piccolo.apps.user.piccolo_app",
    ],
    commands=[],
)



"""
An example of how to configure and run the admin.

Can be run from the command line using `python -m piccolo_admin.example`,
or `admin_demo`.
"""
import os
import sys

from piccolo_api.session_auth.tables import SessionsBase
from piccolo.engine.sqlite import SQLiteEngine
from piccolo.apps.user.tables import BaseUser
from piccolo.table import Table
from piccolo.columns import (
    Varchar,
    Integer,
    ForeignKey,
    Boolean,
    Text,
    Timestamp,
)
from piccolo.columns.readable import Readable

from piccolo_admin.endpoints import create_admin
from piccolo_admin.example_data import DIRECTORS, MOVIES


DB_PATH = os.path.join(os.path.dirname(__file__), "example.sqlite")
DB = SQLiteEngine(path=DB_PATH)


class Sessions(SessionsBase, db=DB):
    pass


class User(BaseUser, db=DB, tablename="piccolo_user"):
    pass


class Director(Table, db=DB):
    name = Varchar(length=300, null=False)

    @classmethod
    def get_readable(cls):
        return Readable(template="%s", columns=[cls.name])


class Movie(Table, db=DB):
    name = Varchar(length=300)
    rating = Integer()
    director = ForeignKey(references=Director)
    won_oscar = Boolean()
    description = Text()
    release_date = Timestamp()


APP = create_admin([Director, Movie], auth_table=User, session_table=Sessions)


def populate_data():
    # Recreate the database
    if os.path.exists(DB_PATH):
        os.unlink(DB_PATH)
    Director.create_table().run_sync()
    Movie.create_table().run_sync()
    User.create_table().run_sync()
    Sessions.create_table().run_sync()

    # Add some rows
    Director.insert(*[Director(**d) for d in DIRECTORS]).run_sync()
    Movie.insert(*[Movie(**m) for m in MOVIES]).run_sync()

    # Create a user for testing login
    user = User(
        username="piccolo",
        password="piccolo123",
        admin=True,
        email="admin@test.com",
    )
    user.save().run_sync()


def main(persist=False, read_only_db=False, use_hypercorn=False):
    """
    If persist is set to True, we don't rebuild all of the data each time.

    If read_only is set to True, the database will be opened in read_only
    mode. Make sure use_hypercorn is also True.
    """
    if not persist:
        populate_data()

    if read_only_db:
        DB.connection_kwargs.update(
            {"uri": True, "database": f"file:{DB_PATH}?mode=ro"}
        )

    # Server
    if use_hypercorn:
        import asyncio
        from hypercorn.asyncio import serve
        from hypercorn.config import Config

        class CustomConfig(Config):
            use_reloader = True

        asyncio.run(serve(APP, CustomConfig()))
    else:
        # Uvicorn (v0.8) doesn't support breakpoints when using auto reload.
        import uvicorn

        uvicorn.run("piccolo_admin.example:APP", reload=True, debug=True)


if __name__ == "__main__":
    args = sys.argv
    persist = "--persist" in args
    read_only_db = "--read-only-db" in args
    use_hypercorn = "--hypercorn" in args
    main(persist=persist, read_only_db=read_only_db)
